/* Sieve of Eratosthenes, based on an old C* example.          */
/* See C* Reference Manual, Version 4.0A Field Test, TMC, 1987 */
/* File:  sieve.cs					       */

#include <stdio.h>
#include <stdlib.h>

/* Size for sieve calculations, 32K is convenient for demos */
#define CSIZE 32768                        /* column size */ 

/* Boolean constants */
#define TRUE  1
#define FALSE 0

/* Other constants */
#define RSIZE 16384                        /* row size */ 
#define DIVISL 29          /* division between small and large */
#define MID_P 16384       /* division for middle and large primes */
#define MAX_P 32768   /* maxinum numbers need to mark off mult. 2^15 */

shape [CSIZE] sieve;	/* Shape for sieve representation */

int:sieve  number;	/* Number represented by this proc. */
bool:sieve is_prime;	/* True if number is prime */
bool:sieve mk[RSIZE];   /* bit vector with primes marked as true */
int:sieve i;            /* Represents pcoord */ 
int:sieve row_i;        /* Number representing #row x pcoord */

int limit;	/* Limit on sieve calculation, < CSIZE */

/* Find all primes <= limit */
/* Assumes  [0]number = x for x = 3,5,7,9, .... */

void find_primes()
 { int next_prime;		/* Next prime identified */
   int nprime;                  /* number of primes */ 
   bool:sieve is_candidate;     /* True of proc's that might be prime */
   with (sieve)
    { /* Initialize */
      is_prime = FALSE;
      is_candidate = (number <= limit);
      next_prime = 1;  /* To start the loop */
      /* Loop continues until all candidates are prime */
      while ( next_prime*next_prime < limit )
       where (is_candidate)
	 { next_prime = (<?= number);
	   where ( (number % next_prime) == 0 ) is_candidate = FALSE;
           where (number == next_prime) is_prime = TRUE;	
        /* Mark next_prime as prime */
	 }
      /* All remaining candidates are prime */
      where (is_candidate) is_prime = TRUE;
      where (number < 32768) nprime = (+= is_prime);
    }  /* end with */
   printf("The number of primes up to 32K is %6d\n",nprime);  
}

/* Print (some of) the primes, 10 per line.		*/
/* pall:  -1 - none,  0 - in last 100 numbers,  1- all	*/

void print_primes(int pall)
 { int j, nx;
   bool:sieve pprime;	/* Identifies primes to be printed */
   j = 1;
   with (sieve)
    { if ( pall >= 0 ) pprime = is_prime; else pprime = FALSE;
      if ( pall == 0 ) pprime &= (number >= (limit - 100));
      /* Loop continues until desired primes are printed */
      /* This amounts to "do-for-selected-processors"	 */
      while (|= pprime)
       where (pprime)
	{ nx = (<?= number);
          where (number == nx) pprime = FALSE;
	  /* Note that nx processed */
	  printf("%6d ", nx);
	  if ( (j++) == 10 ) { printf("\n");  j = 1; };
	} /* end where, end while */
      if (j != 1) printf("\n");
    } /* end with */
 }

/* A bit vector that has the primes marked as true and multiples false. */
/* The small primes counts in parallel to markoff the multiple */
/* of the prime */

void count_markoff(int p)
{ int j;
  int:sieve counter;      /* counter for small primes */
  int:sieve q;
  bool stop=0;

  with (sieve)
   {  
      /* set counter to start count to markoff multiples of p */
     q = row_i - ((p-3)/2);
     counter = (q % p);
     where ((q > 0) & (counter == 0)) counter = p;
     /* loops through marking off multiples of p using count */
     j=0;
     while ((j < RSIZE) && !stop)
      { where (counter == p)
         { mk[j] = FALSE;
           counter = 0;
         } /* end where */
         counter++;
       if ([0]mk[32] == 0) 
         { printf("ERROR\n"); stop++; }  
         j++; 
     } /* end while */
   }  /* end of with */
}        

/* The large primes steps through use indirect address to markoff */
/* the multiple of the prime */

void mid_step_markoff(int p)
{ 
  int k;
  int:sieve stepm;      /* starting address for large primes */ 
  int:sieve qm; 
  bool flag =0;
 
  with (sieve) 
   { 
     /* initialize address of the multiple of p step */
       qm = ((p-3)/2) - row_i;
       stepm = (qm %% p); 
       where (qm >= 0) stepm = stepm + p;
  /* loops through marking off multiples using indirect addressing */
       k=0; 
       while ((k<=(RSIZE-1)/p) && !flag)
       {  where ( stepm < RSIZE)
           { mk[stepm] = FALSE;
             stepm = stepm + p;
            } /* end of where */
          if ([0]mk[32] == 0) 
           { printf("ERROR\n"); flag++; }  
          k++; 
        } /* end of for */ 
   } /* end with */ 
}


/* Prints out at a starting number(num), 25 number and see if it is */   
/* markoff or out */

check_mk(int num)
{
  int col_p,row_p;
  int n,c,tn,j; 

  /* initialization */
  c = 0;
  n = num % 100;
  col_p = (num-3)/(RSIZE * 2);
  row_p = ((num-3)/2) % RSIZE;
  tn = row_p + 25;

  printf("Bits for odd integers starting at %10d \n",num);

  /* printing the numbers that are being checked */  
  for (j=0; j<25; j++)
   { printf("%3d",n);
     if (n == 99) n = -1; 
     n = n + 2; } 
  printf("\n");

   /* looping to print the mark in mk for checking */ 
  for (; row_p < tn; row_p++,c++)
   { 
     if (row_p > (RSIZE-1))
      {  col_p++;
         row_p = 0;
         tn = row_p+(25-c);
         c = 0;
       } /* end if */ 
     printf("%3d",[col_p]mk[row_p]); 
   } /* end for */ 
  printf("\n"); 
}  

/* Main program.  Commands are:
   d <int>  - display status: -1 just time, 0 in last 100, 1 all
   l <int>  - set limit to int
   c <int>  - markoff multiple of <int> by counting
   g <int>  - markoff multiple of <int> by general stepping 
   s <int>  - markoff multiple of <int> primes in 16k...32k stepping  
   k <int>  - starting checking at <int> an odd integer   
   p        - find primes
   q        _ quit           */

void main()
{  int display;  char c[2];
   int cp,gp,n,j; 
   display = 1;  limit = 65535;

    /* Initialize `number' for integers 2, 3, .... */
    with (sieve) number = pcoord(0)*2+1;
    [0]number = 2;

    /* initialize mk,i and row_i */
    with (sieve) 
     { for (j=0; j<RSIZE; j++) mk[j] = TRUE;
        i = pcoord(0);
        row_i = RSIZE * i;  
     } /* end of with */

    while ( printf("sieve>"), scanf("%1s", c), c[0] != 'q' )
     switch (c[0])
      { case 'd': scanf("%d", &display); break;
        case 'l': scanf("%d", &limit);  break;
        case 'p':
	  printf("Calculating primes ...\n");
	  CM_start_timer(1);
		find_primes();
	  CM_stop_timer(1);  CM_reset_timer();
/**             print_primes(display); **/ 
	  break;
        case 'c': scanf("%d", &cp);
          printf("Counting to markoff multiples of %d \n",cp);
          CM_start_timer(1);
                count_markoff(cp);
          CM_stop_timer(1); CM_reset_timer();
          break;
        case 'g': scanf("%d", &gp);
          printf("Stepping to markoff multiples of %d\n",gp);
          CM_start_timer(1);
                mid_step_markoff(gp);
          CM_stop_timer(1); CM_reset_timer();
          break;
        case 'k': scanf("%d",&n);
          check_mk(n);
          break;
	default:   printf("Bad input - ignored\n"); break;
      }
 }







































































